import 'rxjs/add/operator/toPromise';
import { AsyncSubject as AsyncSubject$1 } from 'rxjs/AsyncSubject';
import { Observable as Observable$1 } from 'rxjs/Observable';
import { Inject, InjectionToken, Injector, NgModule, Optional, Pipe } from '@angular/core';
import { BehaviorSubject as BehaviorSubject$1 } from 'rxjs/BehaviorSubject';
import { ReplaySubject as ReplaySubject$1 } from 'rxjs/ReplaySubject';
import { Subject as Subject$1 } from 'rxjs/Subject';
import { CommonModule } from '@angular/common';
import { createStore } from 'redux';
var MetadataManager = /** @class */ (function () {
    function MetadataManager() {
    }
    /**
     * @template T
     * @param {?} target
     * @param {?} data
     * @return {?}
     */
    MetadataManager.setActionMetadata = function (target, data) {
        MetadataManager.set(target, MetadataManager.ACTION_KEY, data);
    };
    /**
     * @template T
     * @param {?} target
     * @return {?}
     */
    MetadataManager.getActionMetadata = function (target) {
        return MetadataManager.get(target, MetadataManager.ACTION_KEY, MetadataManager.ACTION_DEFAULT);
    };
    /**
     * @template T
     * @param {?} target
     * @param {?} data
     * @return {?}
     */
    MetadataManager.setActionContextMetadata = function (target, data) {
        MetadataManager.set(target, MetadataManager.ACTION_CONTEXT_KEY, data);
    };
    /**
     * @template T
     * @param {?} target
     * @return {?}
     */
    MetadataManager.getActionContextMetadata = function (target) {
        return MetadataManager.get(target, MetadataManager.ACTION_CONTEXT_KEY, MetadataManager.ACTION_CONTEXT_DEFAULT);
    };
    /**
     * @template T
     * @param {?} target
     * @param {?} data
     * @return {?}
     */
    MetadataManager.setReducerMetadata = function (target, data) {
        MetadataManager.set(target, MetadataManager.REDUCER_KEY, data);
    };
    /**
     * @template T
     * @param {?} target
     * @return {?}
     */
    MetadataManager.getReducerMetadata = function (target) {
        return MetadataManager.get(target, MetadataManager.REDUCER_KEY, MetadataManager.REDUCER_DEFAULT);
    };
    /**
     * @template T
     * @param {?} target
     * @param {?} data
     * @return {?}
     */
    MetadataManager.setStateMetadata = function (target, data) {
        MetadataManager.set(target, MetadataManager.STATE_KEY, data);
    };
    /**
     * @template T
     * @param {?} target
     * @return {?}
     */
    MetadataManager.getStateMetadata = function (target) {
        return MetadataManager.get(target, MetadataManager.STATE_KEY, MetadataManager.STATE_DEFAULT);
    };
    /**
     * @template D, T
     * @param {?} target
     * @param {?} key
     * @param {?} data
     * @return {?}
     */
    MetadataManager.set = function (target, key, data) {
        Reflect['defineMetadata'](key, data, target);
    };
    /**
     * @template D, T
     * @param {?} target
     * @param {?} key
     * @param {?} defaultData
     * @return {?}
     */
    MetadataManager.get = function (target, key, defaultData) {
        defaultData = defaultData || /** @type {?} */ ({});
        return Object.assign({}, defaultData, Reflect['getMetadata'](key, target));
    };
    return MetadataManager;
}());
MetadataManager.ACTION_KEY = Symbol('@ReduxAction');
MetadataManager.ACTION_DEFAULT = {
    contextClazz: null,
    type: '',
};
MetadataManager.ACTION_CONTEXT_KEY = Symbol('@ReduxActionContext');
MetadataManager.ACTION_CONTEXT_DEFAULT = {
    prefix: '',
};
MetadataManager.REDUCER_KEY = Symbol('@ReduxReducer');
MetadataManager.REDUCER_DEFAULT = {
    reducers: [],
};
MetadataManager.STATE_KEY = Symbol('@ReduxState');
MetadataManager.STATE_DEFAULT = {
    name: '',
};
/**
 * @param {?} config
 * @return {?}
 */
function ReduxActionContextDecorator(config) {
    return function (constructor) {
        MetadataManager.setActionContextMetadata(constructor, config);
        return constructor;
    };
}
var ReduxRegistry = /** @class */ (function () {
    function ReduxRegistry() {
    }
    /**
     * @param {?} store
     * @return {?}
     */
    ReduxRegistry.registerStore = function (store) {
        ReduxRegistry._store.next(store);
        ReduxRegistry._store.complete();
    };
    /**
     * @param {?} stateName
     * @param {?} types
     * @param {?} reducer
     * @return {?}
     */
    ReduxRegistry.registerReducer = function (stateName, types, reducer) {
        if (Array.isArray(types)) {
            types.forEach(function (type) {
                ReduxRegistry._reducers.push({
                    reducer: reducer,
                    stateName: stateName,
                    type: type,
                });
            });
        }
        else {
            ReduxRegistry._reducers.push({
                reducer: reducer,
                stateName: stateName,
                type: /** @type {?} */ (types),
            });
        }
    };
    /**
     * @param {?} state
     * @return {?}
     */
    ReduxRegistry.registerState = function (state) {
        ReduxRegistry.getStore().then(function (store) {
            var /** @type {?} */ stateConfig = MetadataManager.getStateMetadata(state.constructor);
            var /** @type {?} */ initState = state.getInitialState();
            var /** @type {?} */ initStateToResolve = initState;
            if (initState instanceof Observable$1) {
                initStateToResolve = initState.toPromise();
            }
            Promise.resolve(initStateToResolve).then(function (initialState) {
                store.dispatch({
                    payload: {
                        initialState: initialState,
                        stateName: stateConfig.name,
                    },
                    type: ReduxRegistry.ACTION_REGISTER_MODULE,
                });
            });
        });
    };
    /**
     * @param {?} type
     * @return {?}
     */
    ReduxRegistry.getReducerItemsByType = function (type) {
        return ReduxRegistry._reducers
            .filter(function (reducerItem) { return getActionTypeByFunction(reducerItem.type) === type; });
    };
    /**
     * @return {?}
     */
    ReduxRegistry.getStore = function () {
        return new Promise(ReduxRegistry._store.subscribe.bind(ReduxRegistry._store));
    };
    return ReduxRegistry;
}());
ReduxRegistry.ACTION_REGISTER_MODULE = ReduxRegistry.name + "://registerModule";
ReduxRegistry._store = new AsyncSubject$1();
ReduxRegistry._reducers = [];
/**
 * @param {?=} config
 * @return {?}
 */
function ReduxActionDecorator(config) {
    return function (target, propertyKey, descriptor) {
        config = config || { type: propertyKey };
        var /** @type {?} */ originalFunction = descriptor.value;
        var /** @type {?} */ proxyFunction = function () {
            var /** @type {?} */ contextData = MetadataManager.getActionContextMetadata(target.constructor);
            var /** @type {?} */ type = getActionType(contextData.prefix, config.type);
            var /** @type {?} */ returnValue = originalFunction.apply(this, arguments);
            Promise.all([
                Promise.resolve(returnValue),
                ReduxRegistry.getStore(),
            ]).then(function (_a) {
                var payload = _a[0], store = _a[1];
                store.dispatch({ payload: payload, type: type });
            });
            return returnValue;
        };
        MetadataManager.setActionMetadata(proxyFunction, {
            contextClazz: target.constructor,
            type: config.type,
        });
        descriptor.value = proxyFunction;
    };
}
/**
 * @param {?} target
 * @return {?}
 */
function getActionTypeByFunction(target) {
    var _a = MetadataManager.getActionMetadata(target), type = _a.type, contextClazz = _a.contextClazz;
    var prefix = MetadataManager.getActionContextMetadata(contextClazz).prefix;
    return getActionType(prefix, type);
}
/**
 * @param {...?} rest
 * @return {?}
 */
function getActionType() {
    var rest = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        rest[_i] = arguments[_i];
    }
    return rest.filter(function (i) { return i.length > 0; }).join('/');
}
var ReduxModuleRootReducer = /** @class */ (function () {
    function ReduxModuleRootReducer() {
    }
    /**
     * @param {?} state
     * @param {?} action
     * @return {?}
     */
    ReduxModuleRootReducer.reduce = function (state, action) {
        if (action.type === ReduxRegistry.ACTION_REGISTER_MODULE) {
            return ReduxModuleRootReducer.reduceRegisterModule(state, /** @type {?} */ (action));
        }
        return ReduxModuleRootReducer.reduceByRegisteredReducers(state, action);
    };
    /**
     * @param {?} state
     * @param {?} action
     * @return {?}
     */
    ReduxModuleRootReducer.reduceRegisterModule = function (state, action) {
        var _a = action.payload, stateName = _a.stateName, initialState = _a.initialState;
        return Object.assign({}, state, (_b = {},
            _b[stateName] = initialState,
            _b));
        var _b;
    };
    /**
     * @param {?} state
     * @param {?} action
     * @return {?}
     */
    ReduxModuleRootReducer.reduceByRegisteredReducers = function (state, action) {
        var /** @type {?} */ reducers = ReduxRegistry.getReducerItemsByType(action.type);
        return reducers.reduce(function (stateToReduce, reducerItem) {
            var stateName = reducerItem.stateName, reducer = reducerItem.reducer;
            return Object.assign({}, stateToReduce, (_a = {},
                _a[stateName] = reducer(stateToReduce[stateName], action),
                _a));
            var _a;
        }, state);
    };
    return ReduxModuleRootReducer;
}());
/**
 * @template S, P
 * @param {?} types
 * @return {?}
 */
function ReduxReducerDecorator(types) {
    return function (target, propertyKey) {
        var /** @type {?} */ data = MetadataManager.getReducerMetadata(target.constructor);
        data.reducers.push({ types: types, reducer: target[propertyKey].bind(target) });
        MetadataManager.setReducerMetadata(target.constructor, data);
        return target[propertyKey];
    };
}
var REDUX_MODULE_CHILD_CONFIG = new InjectionToken('REDUX_MODULE_CHILD_CONFIG');
var ReduxStateSelectorSubjectType = {};
ReduxStateSelectorSubjectType.DEFAULT = 0;
ReduxStateSelectorSubjectType.SUBJECT = 1;
ReduxStateSelectorSubjectType.BEHAVIOR_SUBJECT = 2;
ReduxStateSelectorSubjectType.REPLAY_SUBJECT = 3;
ReduxStateSelectorSubjectType[ReduxStateSelectorSubjectType.DEFAULT] = "DEFAULT";
ReduxStateSelectorSubjectType[ReduxStateSelectorSubjectType.SUBJECT] = "SUBJECT";
ReduxStateSelectorSubjectType[ReduxStateSelectorSubjectType.BEHAVIOR_SUBJECT] = "BEHAVIOR_SUBJECT";
ReduxStateSelectorSubjectType[ReduxStateSelectorSubjectType.REPLAY_SUBJECT] = "REPLAY_SUBJECT";
var ReduxStateSelector = /** @class */ (function () {
    /**
     * @param {?} expression
     * @param {?=} context
     */
    function ReduxStateSelector(expression, context) {
        this.expression = expression;
        this.context = context;
        if (!expression.startsWith(ReduxStateSelector.DELIMITER)) {
            if (!context) {
                throw new Error('You need to provide a state context if you want to use relativ selectors');
            }
            this.expression = "/" + MetadataManager.getStateMetadata(context).name + "/" + this.expression;
        }
    }
    /**
     * @template S
     * @return {?}
     */
    ReduxStateSelector.prototype.asBehaviorSubject = function () {
        return /** @type {?} */ (this.getBySubjectType(ReduxStateSelectorSubjectType.BEHAVIOR_SUBJECT, null));
    };
    /**
     * @template S
     * @return {?}
     */
    ReduxStateSelector.prototype.asReplaySubject = function () {
        return /** @type {?} */ (this.getBySubjectType(ReduxStateSelectorSubjectType.REPLAY_SUBJECT));
    };
    /**
     * @template S
     * @return {?}
     */
    ReduxStateSelector.prototype.asSubject = function () {
        return /** @type {?} */ (this.getBySubjectType(ReduxStateSelectorSubjectType.SUBJECT));
    };
    /**
     * @template S
     * @param {?} type
     * @param {?=} initialValue
     * @return {?}
     */
    ReduxStateSelector.prototype.getBySubjectType = function (type, initialValue) {
        var _this = this;
        var /** @type {?} */ subject;
        switch (type) {
            case ReduxStateSelectorSubjectType.REPLAY_SUBJECT:
                subject = new ReplaySubject$1();
                break;
            case ReduxStateSelectorSubjectType.SUBJECT:
                subject = new Subject$1();
                break;
            default:
                subject = new BehaviorSubject$1(initialValue || null);
        }
        ReduxRegistry.getStore().then(function (store) {
            subject.next(_this.getValueFromState(store.getState()));
            store.subscribe(function () { return subject.next(_this.getValueFromState(store.getState())); });
        });
        return subject;
    };
    /**
     * @template T, S
     * @param {?} state
     * @return {?}
     */
    ReduxStateSelector.prototype.getValueFromState = function (state) {
        return this.expression.split(ReduxStateSelector.DELIMITER)
            .filter(ReduxStateSelector.isPropertyKeyValid)
            .reduce(function (previousValue, propertyKey) {
            if (!previousValue || !previousValue[propertyKey]) {
                return null;
            }
            return previousValue[propertyKey];
        }, state);
    };
    /**
     * @param {?} propertyKey
     * @return {?}
     */
    ReduxStateSelector.isPropertyKeyValid = function (propertyKey) {
        return propertyKey !== '';
    };
    return ReduxStateSelector;
}());
ReduxStateSelector.DELIMITER = '/';
var ReduxSelectPipe = /** @class */ (function () {
    /**
     * @param {?} moduleConfig
     */
    function ReduxSelectPipe(moduleConfig) {
        this.moduleConfig = moduleConfig;
    }
    /**
     * @param {?} selector
     * @return {?}
     */
    ReduxSelectPipe.prototype.transform = function (selector) {
        return new ReduxStateSelector(selector, this.moduleConfig.state)
            .getBySubjectType(ReduxStateSelectorSubjectType.DEFAULT)
            .asObservable();
    };
    return ReduxSelectPipe;
}());
ReduxSelectPipe.decorators = [
    { type: Pipe, args: [{ name: 'reduxSelect' },] },
];
/**
 * @nocollapse
 */
ReduxSelectPipe.ctorParameters = function () { return [
    { type: undefined, decorators: [{ type: Inject, args: [REDUX_MODULE_CHILD_CONFIG,] },] },
]; };
/**
 * @param {?} expression
 * @param {?=} context
 * @param {?=} subjectType
 * @return {?}
 */
function ReduxSelectDecorator(expression, context, subjectType) {
    return function (target, propertyKey) {
        Object.defineProperty(target, propertyKey, {
            configurable: true,
            enumerable: true,
            value: new ReduxStateSelector(expression, context)
                .getBySubjectType(subjectType, null)
                .asObservable(),
        });
    };
}
/**
 * @param {?} config
 * @return {?}
 */
function ReduxStateDecorator(config) {
    return function (constructor) {
        MetadataManager.setStateMetadata(constructor, config);
        return constructor;
    };
}
var REDUX_MODULE_ROOT_CONFIG = new InjectionToken('REDUX_MODULE_ROOT_CONFIG');
var IS_ROOT_MODULE = new InjectionToken('IS_ROOT_MODULE');
var ReduxModule = /** @class */ (function () {
    /**
     * @param {?=} isRootModule
     * @param {?=} rootConfig
     * @param {?=} childConfig
     * @param {?=} injector
     */
    function ReduxModule(isRootModule, rootConfig, childConfig, injector) {
        if (isRootModule === void 0) { isRootModule = false; }
        if (rootConfig === void 0) { rootConfig = null; }
        if (childConfig === void 0) { childConfig = null; }
        this.injector = injector;
        if (isRootModule) {
            rootConfig = Object.assign({
                store: null,
            }, rootConfig || {});
            ReduxRegistry.registerStore(rootConfig.store || createStore(ReduxModuleRootReducer.reduce, {}));
        }
        if (childConfig) {
            this.initChildConfig(childConfig);
        }
    }
    /**
     * @param {?} childConfig
     * @return {?}
     */
    ReduxModule.prototype.initChildConfig = function (childConfig) {
        if (childConfig.state) {
            ReduxRegistry.registerState(this.injector.get(childConfig.state));
            if (childConfig.reducers) {
                var /** @type {?} */ stateMetadata_1 = MetadataManager.getStateMetadata(childConfig.state);
                childConfig.reducers
                    .map(function (reducer) { return MetadataManager.getReducerMetadata(reducer.constructor); })
                    .forEach(function (metadata) {
                    metadata.reducers.forEach(function (reducer) {
                        ReduxRegistry.registerReducer(stateMetadata_1.name, reducer.types, reducer.reducer);
                    });
                });
            }
        }
    };
    /**
     * @param {?} config
     * @return {?}
     */
    ReduxModule.forChild = function (config) {
        return {
            ngModule: ReduxModule,
            providers: [
                config.state,
                { provide: IS_ROOT_MODULE, useValue: false },
                { provide: REDUX_MODULE_CHILD_CONFIG, useValue: config },
            ],
        };
    };
    /**
     * @param {?=} config
     * @return {?}
     */
    ReduxModule.forRoot = function (config) {
        return {
            ngModule: ReduxModule,
            providers: [
                { provide: IS_ROOT_MODULE, useValue: true },
                { provide: REDUX_MODULE_ROOT_CONFIG, useValue: config },
            ],
        };
    };
    return ReduxModule;
}());
ReduxModule.decorators = [
    { type: NgModule, args: [{
                declarations: [
                    ReduxSelectPipe,
                ],
                exports: [
                    ReduxSelectPipe,
                ],
                imports: [
                    CommonModule,
                ],
            },] },
];
/**
 * @nocollapse
 */
ReduxModule.ctorParameters = function () { return [
    { type: undefined, decorators: [{ type: Inject, args: [IS_ROOT_MODULE,] },] },
    { type: undefined, decorators: [{ type: Optional }, { type: Inject, args: [REDUX_MODULE_ROOT_CONFIG,] },] },
    { type: undefined, decorators: [{ type: Optional }, { type: Inject, args: [REDUX_MODULE_CHILD_CONFIG,] },] },
    { type: Injector, },
]; };
/**
 * Generated bundle index. Do not edit.
 */
export { ReduxActionContextDecorator as ReduxActionContext, ReduxActionDecorator as ReduxAction, ReduxModuleRootReducer, ReduxReducerDecorator as ReduxReducer, ReduxSelectDecorator as ReduxSelect, ReduxSelectPipe, ReduxStateDecorator as ReduxState, ReduxStateSelector, ReduxStateSelectorSubjectType, ReduxModule, REDUX_MODULE_CHILD_CONFIG as ɵa, REDUX_MODULE_ROOT_CONFIG as ɵd, IS_ROOT_MODULE as ɵc };
//# sourceMappingURL=ngx-redux-core.es5.js.map
